/* Ermittelt die Temperatur mit einem Dallas/Maxim DS18S20 (S!)
 * Gibt die Temperatur an einem LCD aus.
 * DS18S20 mit parasitaerer Versorgung.
 * Pinkonfiguration etc. in ds18s20.h
 *
 */

#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h> 		// sprintf
#include "lcd_tools.h"
#include "ds18s20.h"

/**
	@brief	Definiert eine Warteschleife, die Taktzyklen totschlaegt
	@param 	Wartezeit in Schleifendurchlaeufen
			Die Anzahl der Schleifendurchlaeufe ist Abhaengig vom generierten Assemblercode und den benoetigten Ticks
*/
inline __attribute__((gnu_inline)) void ow_delay (uint16_t delay)
{
	while(delay--) asm volatile("nop");
}
 
/**
	@brief	Setzt den Sensor zurueck
	@param 	none
	@return	Erfolg? 0 = OK, 1 = Fehler
*/
uint8_t reset(void)
{
	uint8_t i;
	
	// Datenleitung Low und warten
	LOW();			
	OUTPUT_MODE();
	ow_delay(us(480));
	
	// Datenleitung freigeben
	INPUT_MODE();
	ow_delay(us(60));
	
	// Zustand der Datenleitung ermitteln und Restzeit warten
	i=(PIN & (1 << DQ));
	ow_delay(us(420));
	
	return i;		// Wert der Datennleitung zurueckgeben (0 = OK, 1 = Fehler)
}

/**
	@brief	Schreibt ein Bit auf der Datenleitung
	@param 	Bitwert
	@return	none
*/
void write_bit(uint8_t bit)
{
	// Leitung fuer 1s Low ziehen
	LOW();
	OUTPUT_MODE();
	ow_delay(us(1));
		
	if(bit) 			// Wenn 1 geschrieben werden soll: Leitung freigeben, sonst: bleibt Low
		INPUT_MODE();
		
	// warten und dann Leitung freigeben
	ow_delay(us(60));
	INPUT_MODE();
}

/**
	@brief	Liest ein Bit auf der Datenleitung
	@param 	none
	@return	Bitwert
*/
uint8_t read_bit(void)
{
	uint8_t bit=0;
	
	// Leitung fuer 1s Low ziehen
	LOW();
	OUTPUT_MODE();
	ow_delay(us(1));
	
	// Leitung freigeben und warten
	INPUT_MODE();
	ow_delay(us(14));
	
	if (PIN & (1 << DQ)) 	// Wenn Datenleitung High, dann 1 empfangen
		bit=1;
		
	// Warten und Wert zurueckgeben
	ow_delay(us(45));
	return bit;
}

/**
	@brief	Liest ein Byte ein
	@param 	none
	@return	Byte
*/
uint8_t read_byte(void)
{
	uint8_t i=8, n=0;
	
	while(i--)		// 8x
	{
		n >>= 1;					// Rechtsshift um 1
		n |= (read_bit() << 7);	// Bitwert speichern
	}
	return n;
}

/**
	@brief	Schreibt ein Byte 
	@param 	Byte
	@return	none
*/
void write_byte(uint8_t byte)
{
	uint8_t i=8;
	
	while(i--)		// 8x
	{
		write_bit(byte&1);		// Bit schreiben
		byte >>= 1;				// Rechtsshift um 1
	}
}


/**
	@brief	Temperatur auslesen
	@param 	Array mit Tempwerten
	@return	Erfolg? 0 = OK, 1 = Fehler
*/
uint8_t read_temperature(uint8_t *temperatur)
{
	//Reset, skip ROM and start temperature conversion
	reset();						// Sensor zueruecksetzen
	write_byte(CMD_SKIPROM);		// ROM Adressierung uerbespringen (= alle Sensoren)
	write_byte(CMD_CONVERTTEMP);	// Temperatur ermitteln

	// Paraisitaere Spannungsversorgung fuer Convertierung
	HIGH();							// Datenleitung High
	OUTPUT_MODE();					// als Ausgang
	delay_ms(750);					// t_conf laut Datenblatt
	while(!read_bit());			// Warte bis Temperatur konvertiert wurde

	reset();						// Sensor zueruecksetzen
	write_byte(CMD_SKIPROM);		// ROM Adressierung uerbespringen (= alle Sensoren)
	write_byte(CMD_RSCRATCHPAD);	// Read Scratchpad
	
	temperatur[0]=read_byte();		// 1. Byte einlesen
	temperatur[1]=read_byte();		// 2. Byte einlesen, Rest ignorieren
	return reset();					// Sensor zueruecksetzen und Ergebnis zuerueckgeben
}

/**
	@brief	Temperatur auslesen und realen Wert zurueckgeben
	@param 	none
	@return	Dezimalzahl inkl. Vorzeichen. Bei Fehler: -999
*/
double get_temperature(void)
{
	uint8_t temperatur[2]={0,0};

	if (read_temperature(temperatur))	// Temperatur abfragen. Erfolg?
		return -999;					// Fehler
	
	if (temperatur[1] == 0xFF)	// Vorzeichen Byte gesetzt => negativer Wert
		return (0xFF - temperatur[0] + 1) / 2.0 * -1.0;
	return temperatur[0] / 2.0;
}


int main(void)
{
	char buffer[20];
	double temp;
	
	lcd_ini();		// LCD initialisieren
	
	while(1)		// endlos
	{
		temp = get_temperature();	// Temp abrufen
		if (temp != -999)			// kein Fehler
			sprintf(buffer, "Temp.: %+6.1f %cC", temp, 0xDF);		// 0xDF = Gradzeichen => "C"
		else
		{
			lcd_cls();
			sprintf(buffer, "Messfehler");
		}
		
		lcd_writetext (buffer);
		lcd_gotopos (1,1);
	}
	return 0;
}



