/* Software UART: 8N1
 * Nur Daten empfangen, kein senden. 
 * Empfangene Daten werden in einem Ringpuffer gespeichert
 * 
 * RxD an ICP1 Pin
 */


#include <avr/io.h>
#include <avr/interrupt.h>


#define BAUDRATE 9600
#define RBUFFLEN 40 //Pufferlnge fr seriellen Empfang


// ICP1 Eingang festlegen:
#define SUART_RXD_PORT PORTD
#define SUART_RXD_PIN  PIND
#define SUART_RXD_DDR  DDRD
#define SUART_RXD_BIT  PD6

static volatile uint16_t inframe;
static volatile uint8_t inbits, received;
static volatile uint8_t indata;
volatile unsigned char rbuff[RBUFFLEN];		// Ringpuffer
volatile uint8_t 	rbuffpos, 					// Position, die als nchstes gelesen werden mu im Ringpuffer
					rbuffcnt; 					// Anzahl zu lesender Zeichen im Puffer

// Initialisierung fr einen ATmega8 
// Fr andere AVR-Derivate sieht dies vermutlich anders aus: 
// Registernamen ndern sich (zB TIMSK0 anstatt TIMSK, etc). 
void softuart_init(void)
{
    cli();

    // Mode #4 fr Timer1 
    // und volle MCU clock 
    // IC Noise Cancel 
    // IC on Falling Edge 
    TCCR1A = 0;
    TCCR1B = (1 << WGM12) | (1 << CS10) | (0 << ICES1) | (1 << ICNC1);

    OCR1A = (uint16_t) ((uint32_t) F_CPU/BAUDRATE);	// OutputCompare fr gewnschte Timer1 Frequenz in Anbhaengigkeit von Baudrate

    SUART_RXD_DDR  &= ~(1 << SUART_RXD_BIT);			// Eingangspin (ICP1) als Eingang
    SUART_RXD_PORT |=  (1 << SUART_RXD_BIT);			// Pull-Up
    TIMSK |= (1 << TICIE1);							// Input Capture Interrupt Enable (Start-Bit Erkennung)
    TIFR  |= (1 << ICF1) | (1 << OCF1B);				// Timer/Counter1, Input Capture Flag und Timer/Counter1, Output Compare B Match Flag

	sei();
}

// reagiere auf Start-Bit (ICP1)
SIGNAL (SIG_INPUT_CAPTURE1)
{
    uint16_t icr1  = ICR1;					// Zeitpunkt des IRQ (TCT1) steht in ICR1: merken
    uint16_t ocr1a = OCR1A;					// Output Compare Register A Einstellung merken
   
    uint16_t ocr1b = icr1 + ocr1a/2;		// Eine halbe Bitzeit zu ICR1 addieren (modulo OCR1A) und nach OCR1B
    if (ocr1b >= ocr1a)
        ocr1b -= ocr1a;
    OCR1B = ocr1b;							// Output Compare Registers B setzen
   
    TIFR = (1 << OCF1B);					// IRQ Output Compare Registers B aktivieren
    TIMSK = (TIMSK & ~(1 << TICIE1)) | (1 << OCIE1B);		// Input Capture Interrupt Enable umschalten (Start-Bit Ueberwachung aus), Output Compare B Match Interrupt Enable
    inframe = 0;							// neues Byte startet
    inbits = 0;								// noch keine Bits empfangen
}

// im Takt der Baudrate 
SIGNAL (SIG_OUTPUT_COMPARE1B)
{
    uint16_t data = inframe >> 1;					// Bereits empfangene Bits eins nach rechts (LSB kommt zuerst)
   
    if (SUART_RXD_PIN & (1 << SUART_RXD_BIT))		// Wenn High am Eingang, dann 1 an 10. Stelle speichern (Shift links um 9)
        data |= (1 << 9);
      
    uint8_t bits = inbits+1;						// Zaehler fuer Anzahl der empfangenen Bits +1
   
    if (10 == bits)									// Wenn 10 Bits empfangen wurden
    {
        if ((data & 1) == 0)						// Wenn das 1. empfangene Bit (Start-Bit) 0 ist, dann wurde gueltiges Byte empfangen
		{
            if (data >= (1 << 9))					// Wenn 10. Bit (Stop-Bit) und irgendwelche Daten 1 ist, dann wurde komplettes Byte empfangen
            {
                indata = data >> 1;					// Byte durch Linksshift vom Start-Bit befreien und durch cast auf uint8_t auch von Parity und Stop-Bit befreien
				if(rbuffcnt < RBUFFLEN)				// keinen Zeichen in einem vollen Ringpuffer berschreiben
					rbuff[(rbuffpos+rbuffcnt++) % RBUFFLEN] = indata;	// welche Position? Gelesene Zeichenpos + Anzahl Zeichen MODULO Pufferlnge 
                received = 1;						// Flag setzen
            }
		}
        TIMSK = (TIMSK & ~(1 << OCIE1B)) | (1 << TICIE1);		// Input Capture Interrupt Enable umschalten (Start-Bit Ueberwachung ein), Output Compare B Match Interrupt Enable umschalten (aus)
        TIFR = (1 << ICF1);						// auf ICP1 reagieren
    }
    else
    {
        inbits = bits;
        inframe = data;
    }
}

// Wartet auf den Empfang eines Zeichens und gibt dieses dann zurueck
unsigned char softuart_getc(void)
{
	unsigned char c;

	while(!rbuffcnt);   		// Warte bis ein Zeichen vorhanden ist
	
	cli();						// Interruptbehandlung kurz aussetzen. Ab jetzt mu es schnell gehen (wenig Befehle), damit Zeichen, die 
								// ab jetzt eintreffen nicht verloren gehen.
	rbuffcnt--;					// anschl. ein Zeichen weniger zum ausgeben
	c = rbuff [rbuffpos++];	// Zeichen holen, was nach dem bereits gelesenen liegt
	if (rbuffpos >= RBUFFLEN)  rbuffpos = 0;	// wenn hinterstes Zeichen (rechts im Puffer) gelesen wurde, dann wieder vorne anfangen

	sei();						// Interruptbehandlung wieder aktivieren

	return (c);					// Zeichen zurckgeben
}


