/* IC (TWI) Bus
 *
 * Speichert Werte in einem externen EEPROM
 *
 * Liest die Werte wieder aus und zeigt sie per UART an
 *
 */

#include <avr/io.h>
#include <util/delay.h>

#define BASEADR		0xA0	// Basisadresse des EEPROMs 24C16, Page 0
#define TW_READ		1		// LSB bei Busadresse zur Kennzeichnung eines Lesezugriffes
#define TW_WRITE	0		// LSB bei Busadresse zur Kennzeichnung eines Schreibzugriffes

#define UART_BAUD_RATE 9600L
#define UART_BAUD_CALC(UART_BAUD_RATE,F_CPU) ((F_CPU)/((UART_BAUD_RATE)*16L)-1)

void uart_putc(unsigned char);
void i2c_start(void);
void i2c_stop(void);
void i2c_send(uint8_t) ;
uint8_t i2c_receive(uint8_t);
void EEPROM_ByteWrite(uint16_t, uint8_t);
uint8_t EEPROM_ByteRead(uint16_t);
int main(void);

// Ein Zeichen senden
void uart_putc(unsigned char c)
{
    while(!(UCSRA & (1 << UDRE)));   	// warte, bis UDR bereit 
    UDR = c;    						// sende Zeichen 
}

/**
	@brief	Startanweisung senden
	@param	none
	@return	none
*/
void i2c_start(void) 
{
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN); 	// send start condition
	while ((TWCR & _BV(TWINT)) == 0) ; 				// wait for transmission
}

/**
	@brief	Stopanweisung senden
	@param	none
	@return	none
*/
void i2c_stop(void) 
{
	TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);		// send stop condition
}

/**
	@brief	Datenbyte senden
	@param	Byte
	@return	none
*/
void i2c_send(uint8_t DataByte) 
{
	TWDR = DataByte;
	TWCR = _BV(TWINT) | _BV(TWEN); 			// clear interrupt to start transmission
	while ((TWCR & _BV(TWINT)) == 0) ; 		// wait for transmission
}

/**
	@brief	Datenbyte empfangen
	@param	ACK senden (1) oder nicht (0)
	@return	none
*/
uint8_t i2c_receive(uint8_t ack) 
{
	if (ack == 1) 
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
	else 
		TWCR = _BV(TWINT) | _BV(TWEN);  		// clear interrupt to start transmission
	while ((TWCR & _BV(TWINT)) == 0) ;   		// wait for transmission
	return TWDR; 
}

/**
	@brief	Byte im EEPROM an Adresse speichern
	@param	Adresse, Datenbyte
	@return	none
*/
void EEPROM_ByteWrite(uint16_t addr, uint8_t wert) 
{
again:
	i2c_start();
	i2c_send(BASEADR | TW_WRITE);

	// Pruefen, ob Adressbyte angenommen wurde
	if ((TWSR & 0xF8) != 0x18)
	{
		i2c_stop();
		goto again;
	}

	//  i2c_send(addr >> 8, 0);		// bei 16-Bit Adressierung HSB zuerst
	i2c_send(addr);

	i2c_send(wert);
	i2c_stop();
	_delay_ms(5);			// warten fuer 24Cxx
}

/**
	@brief	Byte aus EEPROM von Adresse lesen
	@param	Adresse
	@return	Datenbyte
*/
uint8_t EEPROM_ByteRead(uint16_t addr) 
{
	uint8_t wert;
	
	i2c_start();
	i2c_send(BASEADR | TW_WRITE);

	//  i2c_send(addr >> 8, 0);		// bei 16-Bit Adressierung HSB zuerst
	i2c_send(addr);

	i2c_start();
	i2c_send(BASEADR | TW_READ);
	wert = i2c_receive(0);
	i2c_stop();
	return wert;
}

int main(void)
{
	uint8_t wert=0;
	
	UCSRB |= (1 << TXEN);										// UART TX (senden) einschalten
	UCSRC |= (1<<URSEL)|(1<<UCSZ0)|(1<<UCSZ1);				// Asynchron, 8N1
	UBRRH=(uint8_t)(UART_BAUD_CALC(UART_BAUD_RATE,F_CPU)>>8);	// Baudrate whlen
	UBRRL=(uint8_t)UART_BAUD_CALC(UART_BAUD_RATE,F_CPU);

	TWBR = (F_CPU / 100000UL - 16) / 2;	// TWI Baudrate 100KHz

	EEPROM_ByteWrite(0,65);			// Wert im EEPROM an Adresse 0 (Page 0) speichern
	EEPROM_ByteWrite(1,66);			// Wert im EEPROM an Adresse 1 (Page 0) speichern
	EEPROM_ByteWrite(2,67);			// Wert im EEPROM an Adresse 1 (Page 0) speichern

	wert = EEPROM_ByteRead(0);		// Wert aus EEPROM Adresse 0 (Page 0) auslesen
	uart_putc (wert);				// Wert an UART senden
	
	wert = EEPROM_ByteRead(1);		// Wert aus EEPROM Adresse 1 (Page 0) auslesen
	uart_putc (wert);				// Wert an UART senden

	wert = EEPROM_ByteRead(2);		// Wert aus EEPROM Adresse 2 (Page 0) auslesen
	uart_putc (wert);				// Wert an UART senden

	return 0;
}
