/* Temperaturmessung mit einem LM 335 an ADC1
 *
 * Am ATmgea16 wird ein LM335 Temperatursensor angeschlossen zur Demonstration des internen A/D-Wandlers.
 * Setzt voraus, da die externe Referenzspannung AREF angeschlossen ist
 *
 * -----
 *     | ADC1
 *   39|-------------------*-------                             ----\  LM 335 bottom view
 *     |                   |      |2                            |    \ 
 *     | AVCC  ______      |   |__|__                           | o 1 \  
 *   30|------|______|-----|     / \ |  LM 335                  | o 2  |         
 *     |       1K Ohm           /___\                           | o 3 /
 *     |                          |                             |    /
 *     |                          |1                            ----/    
 *     |                         _|_ 
 *     |ATmega16                     GND
 * -----
 *
 *
 * 1 Temperaturnderung (Celsius/Fahrenheit) = 0.01V nderung
 *
 * Als Referenz wird die gemessene Spannung bei Zimmertemperatur definiert. Dieser Wert mu angepat werden.
 *
 */

#include "lcd_tools.h"
#include <stdlib.h>
#include <avr/io.h>

#define TEMP 22.5
#define V_TEMP 2.969					// Spannungsreferenz bei TEMP 
#define V_0GRAD (V_TEMP-(TEMP/100.0))	// Spannung bei Gefrierpunkt

void main()
{
	char 		texte[10];
	double 		komma;
	uint8_t		i;
	uint16_t	result = 0;			

	lcd_ini();

	ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);    	// Frequenzvorteiler: setzen auf 64 (8 MHz / 64 = 125 kHz) und ADC aktivieren 
	ADMUX = 0x01;				// Kanal waehlen (ADC1)
	ADMUX |= (1<<REFS0); 		// externe Referenzspannung nutzen 

	ADCSRA |= (1<<ADSC);              	// eine ADC-Wandlung 
	while ( ADCSRA & (1<<ADSC) );     	// auf Abschluss der Konvertierung warten 
	result = ADCW;						// Wert abholen
	
	result = 0;
	while (1)
	{
		// Eigentliche Messung - Mittelwert aus 4  Wandlungen 
		for(i=0;i<4;i++)
		{
			ADCSRA |= (1<<ADSC);				// eine Wandlung "single conversion"
			while ( ADCSRA & (1<<ADSC) );   	// auf Abschluss der Konvertierung warten
			result += ADCW;						// Wandlungsergebnisse aufaddieren
		}
		result /= 4;		// Summe durch vier teilen = arithm. Mittelwert
	
		lcd_gotoline (1);
		lcd_writetext ("A/D-Wert  : ");
		dtostrf ((double)result, 4, 0, texte);
		lcd_writetext (texte);
		
		lcd_gotoline (2);
		komma = result * 5.0 / 1024.0;			// A/D-Wert in Spannung umrechnen (AREF=5V, 10 Bit A/D-Wandler: 2^10=1024)
		dtostrf (komma, 5, 3, texte);
		lcd_writetext ("Spannung  : ");
		lcd_writetext (texte);
		lcd_writechar ('V');
		
		lcd_gotoline (3);
		komma = komma - V_0GRAD;			// akt. Spannung - Spannung Gefrierpunkt
		komma *= 100.0;						// in C umrechnen
		dtostrf (komma, 6, 1, texte);
		lcd_writetext ("Temperatur: ");
		lcd_writetext (texte);
		lcd_writetext ("C");		

		delay_ms (1000);	// 1 Sekunde  warten. So schnell ndert sich eh nichts
		result=0;
	}	

	// wird nie erreicht:
	ADCSRA &= ~(1<<ADEN);             // ADC deaktivieren 
	while(1);
}



