/* IC (TWI) Bus
 *
 * Portexpander mit PCF8574AP
 *
 * Laesst eine LED an Pin 2 des Expanders leuchten
 * Reagiert auf Taster an Pin 6 und laesst dann LED2 auf Evaluationsboard leuchten
 */

#include <avr/io.h>
#include <util/delay.h>

#define BASEADR		0x76	// Basisadresse des PCF8574AP 0x70 + (A2=0, A1=A0=1)
#define TW_READ		1		// LSB bei Busadresse zur Kennzeichnung eines Lesezugriffes
#define TW_WRITE	0		// LSB bei Busadresse zur Kennzeichnung eines Schreibzugriffes

void i2c_start(void);
void i2c_stop(void);
void i2c_send(uint8_t) ;
uint8_t i2c_receive(uint8_t);
void set_port(uint8_t);
uint8_t get_port(void);
int main(void);

/**
	@brief	Startanweisung senden
	@param	none
	@return	none
*/
void i2c_start(void) 
{
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN); 	// send start condition
	while ((TWCR & _BV(TWINT)) == 0) ; 			// wait for transmission
}

/**
	@brief	Stopanweisung senden
	@param	none
	@return	none
*/
void i2c_stop(void) 
{
	TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);	// send stop condition
}

/**
	@brief	Datenbyte senden
	@param	Byte
	@return	none
*/
void i2c_send(uint8_t DataByte) 
{
	TWDR = DataByte;
	TWCR = _BV(TWINT) | _BV(TWEN); 			// clear interrupt to start transmission
	while ((TWCR & _BV(TWINT)) == 0) ; 		// wait for transmission (ACK)
}

/**
	@brief	Datenbyte empfangen
	@param	ACK senden (1) oder nicht (0)
	@return	none
*/
uint8_t i2c_receive(uint8_t ack) 
{
	if (ack == 1) 
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
	else 
		TWCR = _BV(TWINT) | _BV(TWEN);  		// clear interrupt to start transmission
	while (!(TWCR & (1<<TWINT)));	   			// wait for transmission
	return TWDR; 
}

/**
	@brief	Sendet Daten an PCF8574AP => Ausgnge setzen
	@param	Datenbyte (Portzustand)
	@return	none
*/
void set_port(uint8_t wert) 
{
	i2c_start();
	i2c_send(BASEADR | TW_WRITE);	// Adresse Bausteins: schreibender Zugriff erfolgt
	i2c_send(wert);					// Zustand Pins aendern
	i2c_stop();
}

/**
	@brief	Byte von PCF8574AP einlesen
	@param	none
	@return	Datenbyte
*/
uint8_t get_port(void) 
{
	uint8_t wert;
	
	i2c_start();
	i2c_send(BASEADR | TW_READ);	// Adresse Bausteins: lesender Zugriff erfolgt
	wert = i2c_receive(1);			// Pins abfragen (mit ACK)
	i2c_stop();
	return wert;
}

int main(void)
{
	uint8_t wert=0;
	
	DDRD |= (1 << PD6);					// LED2 an PD6
	
	TWBR = (F_CPU / 100000UL - 16) / 2;	// TWI Baudrate 100KHz

	set_port (0b01000000);					// P2 Low = LED leuchtet, andere High = Eingang
		
	while (1)								// endlos
	{
		if (get_port() & 0b01000000)		// pruefe, ob P6 auf High liegt (Taster ist Low Aktiv)
			PORTD &= ~(1 << PD6);			// LED aus
		else
			PORTD |= (1 << PD6);			// LED an
	}
	
	return 0;
}
