/** Nutzt eine Real Time Clock vom Typ DS1302
 *
 *	Daten werden an die serielle Schnittstelle gesendet (9600, 8N1)
 *
 *	Es wird kein Burst Mode unterstuetzt.
 *	Es wird im 24h Format gearbeitet
 *	Es erfolgt keine Ueberwachung, ob das Datum etc. korrekt ist. Dies kann zum Beispiel
 *	geschehen, in dem ein Wert im RAM gespeichert wird. Ist dieser nicht vorhanden, dann
 *	hat die RTC die Spannung verloren und die Zeit ist ungueltig.
 *
 */

#include <avr/io.h>
#include <stdio.h>
#include <uart.h>
#include <util/delay.h>

#include "rtc_ds1302.h"

#define UHRZEITSETZEN		// Zeile auskommentieren, wenn Uhrzeit nicht gesetzt werden soll

/**
	@brief	Startphase generieren
	@param	none
	@return	none
*/
void DS1302_start(void)
{
	CE_LOW();				// RTC not enabled
	CLK_LOW();				// Takt
	CE_HIGH();				// RTC enable
	_delay_us(4);			// tCC = 4 s
}

/**
	@brief	Abschlussphase generieren
	@param	none
	@return	none
*/
void DS1302_stop(void)
{
	CE_LOW();				// RTC not enabled
	_delay_us(4);			// tCWH = 4 s
}

/**
	@brief	Hilfsfunktion zum Schreiben eines Bytes. Erzeugt den Takt und schreibt Bits
	@param	Bytewert
			Folgt danach ein Lesezugriff? 1=ja, 0=nein
	@return	none
*/
void DS1302_togglewrite(uint8_t byte, uint8_t release)
{
	int i;

	DATA_OUTPUT_MODE();				// Daten an RTC senden

	for(i = 0; i < 8; i++)			// Alle 8 Bits des Byte von LSB nach HSB durchlaufen
	{ 
		if ((byte >> i) & 0x01)	// Wenn Bit 1 ist => HIGH
			DATA_HIGH();
		else
			DATA_LOW();
		
		_delay_us (1);				// tDC = 200 ns = 0,2 s

		CLK_HIGH();					// Clock High = Bit liegt an => uebernehmen
		_delay_us (1);				// tCH = 1000 ns = 1 s

		if( release && i == 7)		// Anschliessend sendet der RTC Daten. Also Leitung als Eingang. 
			DATA_INPUT_MODE();	
		else
		{
			CLK_LOW();				// Clock Low = Bit fertig
			_delay_us (1);			// tCL = 1000 ns = 1 s
		}
	}
}

/**
	@brief	Hilfsfunktion zum Lesen eines Bytes. Erzeugt den Takt und liest Bits
	@param	none
	@return	Bytewert
*/
uint8_t DS1302_toggleread(void)
{
	uint8_t i, data=0;

	for(i = 0; i < 8; i++)			// alle 8 Bits einlesen (LSB first)
	{
		CLK_HIGH();					// Clock High = Vorbereiten
		_delay_us (1);				// tCCZ = 280 ns = 0,28 s
		CLK_LOW();					// Daten anlegen
		_delay_us (1);				// tCCD = 800 ns =0,8 s
		
		if (DATA_LEVEL())			// Wenn High am Eingang => Bit setzen
			data |= (1 << i);
	}
	return( data);
}

/**
	@brief	Liest ein Byte eines Registers ein
	@param	Registeradresse die gelesen werden soll
	@return	Bytewert
*/
uint8_t DS1302_readValue(uint8_t adresse)
{
	DS1302_start();
	adresse += 1;						// Aus Registeradresse fuer Schreibzugriff Adresse fuer Lesezugriff berechnen
	DS1302_togglewrite(adresse, 1);		// Sende anzusprechende Adresse. Anschliessend wird gelesen
	adresse = DS1302_toggleread();		// Byte (Registerinhalt) einlesen
	DS1302_stop();
	return adresse;
}

/**
	@brief	Schreibt ein Byte in ein Registers 
	@param	Registeradresse die beschrieben werden soll
			Bytewert
	@return	none
*/
void DS1302_writeValue(uint8_t adresse, uint8_t byte)
{
	DS1302_start();		
	DS1302_togglewrite(adresse, 0);		// Sende anzusprechende Adresse. Anschliessend wird geschrieben
	DS1302_togglewrite(byte, 0);		// Sende Wert fuer die Adresse.
	DS1302_stop();
}

/**
	@brief	Schreibt die Zeit in die RTC
	@param	Zeit
	@return	none
*/
void WriteZeit(Trealtime zeit)
{
	uint8_t i;

	DS1302_writeValue(DS1302_ENABLE, 0);	// Schreibzugriffe auf RTC erlauben
	DS1302_writeValue(DS1302_SECONDS, 1);	// RTC anhalten
	DS1302_writeValue(DS1302_HOURS, 0);		// 24h Modus
	
	i = (((uint8_t)(zeit.date / 10)) << 4) | (zeit.date % 10);	// Dezimal nach BCD
	DS1302_writeValue(DS1302_DATE, i);		

	i = (((uint8_t)(zeit.month / 10)) << 4) | (zeit.month % 10);	
	DS1302_writeValue(DS1302_MONTH, i);		

	i = (((uint8_t)(zeit.year / 10)) << 4) | (zeit.year % 10);	
	DS1302_writeValue(DS1302_YEAR, i);		

	DS1302_writeValue(DS1302_DAY, zeit.dow);		

	i = (((uint8_t)(zeit.hour / 10)) << 4) | (zeit.hour % 10);	
	DS1302_writeValue(DS1302_HOURS, i);		

	i = (((uint8_t)(zeit.min / 10)) << 4) | (zeit.min % 10);		
	DS1302_writeValue(DS1302_MINUTES, i);		

	i = (((uint8_t)(zeit.sec / 10)) << 4) | (zeit.sec % 10);		
	DS1302_writeValue(DS1302_SECONDS, i);	// Sekunden schreiben und RTC starten (8. Bit = 0)
}

/**
	@brief	Liest die Zeit aus der RTC
	@param	none
	@return	Zeit
*/
Trealtime ReadZeit(void)
{
	Trealtime zeit;
	uint8_t i;

	i = DS1302_readValue(DS1302_DATE);
	i = ((i >> 4) * 10) + (i & 0x0F);		// BCD nach Dezimal
	zeit.date = i;

	i = DS1302_readValue(DS1302_MONTH);
	i = ((i >> 4) * 10) + (i & 0x0F);
	zeit.month = i;

	i = DS1302_readValue(DS1302_YEAR);
	i = ((i >> 4) * 10) + (i & 0x0F);
	zeit.year = i;

	i = DS1302_readValue(DS1302_DAY);
	zeit.dow = i;

	i = DS1302_readValue(DS1302_SECONDS);
	i &=  0x7F;								// CH Flag entfernen
	i = ((i >> 4) * 10) + (i & 0x0F);
	zeit.sec = i;

	i = DS1302_readValue(DS1302_MINUTES);
	i = ((i >> 4) * 10) + (i & 0x0F);
	zeit.min = i;

	i = DS1302_readValue(DS1302_HOURS);
	i &=  0x7F;								// 12/24 Flag entfernen
	i = ((i >> 4) * 10) + (i & 0x0F);
	zeit.hour = i;

	return zeit;
}

/**
	@brief	Hauptprogramm
*/
int main(void)
{
	Trealtime zeit;
	char buf[40];			// Platz fuer alle zeichen der Ausgabe vorsehen

	CLK_OUTPUT_MODE();		// Datenleitungen einstellen
	CE_OUTPUT_MODE();
	DATA_OUTPUT_MODE();	

	uart_ini();				// UART einrichten

#ifdef UHRZEITSETZEN		// nur Ausfuehren, wenn die Uhrzeit gesetzt werden soll. Wird am Anfang des Codes definiert
	zeit.sec = 49;
	zeit.min = 58;
	zeit.hour = 23;			// 23:58:49
	zeit.date = 31;
	zeit.month = 12;
	zeit.year = 14;			// 31.12.14
	zeit.dow = 3;			// Mittwoch
	WriteZeit (zeit);
#endif

	while (1)				// Endlos Zeit auslesen und anzeigen
	{
		zeit = ReadZeit();

		sprintf (buf, "%s, %02d.%02d.20%02d - %02d:%02d:%02d\r", wochentag[zeit.dow-1], zeit.date, zeit.month, zeit.year, zeit.hour, zeit.min, zeit.sec);
		uart_puts(buf);
		
		_delay_ms(1000);
	};
}
